# Description: This script creates a dictionary of motifs from a 
#              index file, reads in a flux library file and selects
#              fragments that overlap a motif. These fragments are 
#              used as a new library.

# Contact: wanja.kassuhn@web.de

# imports
import sys
import random
from collections import defaultdict

# init
motifs = defaultdict(list)

# open index file
index = open(sys.argv[1], 'r')

# store read length
readlength = int(sys.argv[3])

#store binding site affinity
affinity=1.0
affinity = float(sys.argv[4])

# iterate index file
for line in index:
    
    # skip head line and empty
    if not line.startswith('#') and not line in ['\n', '\r\n']:
        
        # split line
        sep = line.split('\t')
        
        # extract transcript ID
        transcriptID = sep[1].split('|')[1]

        # only add transcript IDs once to the dictionary
        if not transcriptID in motifs:
            
            # fill dictionary with transcript IDs
            motifs[transcriptID] = [sep[2],sep[3]]
        
        # add multiple motifs to dictionary    
        else:
            
            motifs[transcriptID].append(sep[2])
            motifs[transcriptID].append(sep[3])
            
# close index file
index.close()

# open library file
lib = open(sys.argv[2], 'r')

# iterate library file
for line in lib:
    
    # split line of library file
    sep = line.split('\t')
    
    # extract transcript ID and fragment start and stop coordinate
    transcriptID = sep[2].split('@')[1].split(".")[0]
    start = int(sep[0])
    stop = int(sep[1])
    
    # check if fragment overlaps a motif
    if transcriptID in motifs:
        #print 1
        # check each motifs start and stop coordinate
        for i in range(0,len(motifs[transcriptID]),2):

            # check for overlaps and print line 
            if start < int(motifs[transcriptID][i]) and stop > int(motifs[transcriptID][i+1]):
                if int(stop) - int(start) <= int( float(readlength) + (float(readlength)/2.0)) and int(stop) - int(start) >= readlength:
                    
                    # decreasing the selected fragments according to user set affinity
                    numFrag=int(float(int(sep[3])*10) * affinity)
                    sys.stdout.write(sep[0] + "\t" + sep[1] + "\t" + sep[2] + "\t" + str(numFrag) + "\n")

# close library file
lib.close() 

# empty dictionary
motifs.clear()   
    
exit
