# Description: This script is used to calculate the genomic
#              coordinates of motifs found in the transcriptome.
#              The transcriptome has to follow a specific  
#              structure in order to work with this script. The script 
#              reads in the output of FIMO and creates a dictionary
#              of all found motifs. The transcriptomic positions are than
#              converted to genomic coordinates (with help of exon
#              coordinates and order).

# Contact: wanja.kassuhn@web.de

# import
import sys
import fileinput
import os.path
from collections import defaultdict

numArgs=len(sys.argv)

if numArgs <= 2:
	sys.exit("Not enough arguments!")

InputDir = sys.argv[1]
Input2= sys.argv[2]
motiflength=int(sys.argv[3])

# open the input file
if os.path.isfile(InputDir+"/transcripts.tar.gz"):
	import tarfile
	transcripts = tarfile.open(InputDir+"/transcripts.tar.gz").extractfile("transcripts.fasta")
else:
	if os.path.isfile(InputDir+"/transcripts.fasta"):
		transcripts = open(InputDir+"/transcripts.fasta" , 'r')
	else:
		sys.stderr.write('ERROR: Could not find "transcripts.tar.gz" or "transcripts.fasta" in "%s".\n' % InputDir)
		sys.exit()
		
# motif file
motif=iter(fileinput.input([Input2]))
# skip header
next(motif)

# create dictionary of motifs
motifs=defaultdict(list)

# iterate through motifs
for line in motif:
	
	# split the motif into single information
	vec=line.split("\t")
	#q-value < 1
	
	if float(vec[7]) < 1.0:

		# in case of multiple motifs in one gene ( same key )
		if vec[1] in motifs:
			motifs[vec[1]].append(vec[2])
			motifs[vec[1]].append(vec[3])
		#adding motif to dictionary
		else:
			#create list of motif start end per gene
			liste=[vec[2],vec[3]]
			helper={ vec[1] : liste }
			motifs.update(helper)
			helper.clear()

#debugging
#print motifs
#print transcripts


# iterate through genes
for line2 in transcripts:
	
	# if the current line is a header
	if line2.startswith('>'):
		
		# separate information in strings and cut the ">" from name of gene in the header
		vector=line2[1:].split(":")
		
		#if motifs on current gene exist
		if vector[0] in motifs:
			
			# number of start- and end positions in the header of current gene on which a motif was found ( -5 for id's and overall position of gene)
			numExon= (len(vector) -5)/2 
			
			# list of exon length
			Exons=[]
			# anchor ( first exon end - start for length )
			Exons.append(int(vector[6]) - int(vector[5]) +1)
			
			# in case exon is smaller than motif
			if (int(vector[6]) - int(vector[5])) < motiflength:
				# Warning
				#print ("This exon on gene %s is smaller than the motif \t\t" % vector[0])
				pass
			
			#debugging
			#print vector[6] + "  -  " + vector[5]
			
			# calculating the end position of each exon on the transcript
			for count in range(1,numExon):
				Exons.append(Exons[count - 1] + int(vector[6+(count *2)]) - int(vector[5+(count *2)]) +1)
				
				# in case exon is smaller than motif
				if (int(vector[6+(count *2)]) - int(vector[5+(count *2)])) < motiflength:
					# Warning
					# print ("This exon on gene %s is smaller than the motif \t\t" % vector[0])
					pass
				
				#debugguing
				#print vector[2*count + 6] + "   -   " + vector[2*count + 5]
			#print vector
			#print Exons
			#print motifs[vector[0]]
			
			# getting list size in case there are multiple motifs in a gene => list size > 1
			i=len(motifs[vector[0]]) / 2
			
			# if there is one motif
			for c in range(i):		
				
				# motif start
				s=int(motifs[vector[0]][2*c])
				# begin of motif on genome
				begin=0
				# end of motif on genome
				end=0
				# end if motif overlaps a splice site
				es=0
				# motif end
				e=int(motifs[vector[0]][2*c + 1])
				# start if motif overlaps a splice site
				ss=0
				
				# iterate over exons of gene
				for c2 in range(len(Exons)):
					
					# look if motif is on first exon for the whole motif on exon
					
					#INFORMATION
					#  vector[5+2*c2]  is the genomic start position of current exon
					#  vector[6+2*c2]  is the genomic end position of current exon
					#  vector[5+2*(c2+1)] is the genomic start position of the next exon
					#  vector[6+2*(c2+1)] is the genomic end position of the next exon ( not used )
					#  Exons[c2] is the length of all previous and the current exon added together
					#  Exons[c2-1] is the length of all previous exons added together
					
					if vector[4] == "-1":
					
						# in case motif is not on exon 1
						if s < (Exons[c2] - motiflength) and c2 != 0:
							begin=int(vector[6+2*c2]) - ( s - Exons[c2 -1] ) +1
							end=int(vector[6+2*c2]) - ( e - Exons[c2 -1] )  +1
							break
						
						# motif on exon 1
						elif s < (Exons[c2] - motiflength):
							begin=int(vector[6+2*c2]) - s  +1
							end=int(vector[6+2*c2]) - e   +1
							break
						
						# motif overlaps a splice site
						# motif not on the first exon or last exon ( makes no sense but apparently happens)
						elif s < Exons[c2] and c2 != 0 and c2 != (len(Exons) -1):
							begin=int(vector[6+2*c2]) - ( s - Exons[c2 -1] )  +1
							es=int(vector[5+2*c2])
							ss=int(vector[6+2*(c2+1)])
							end=int(vector[6+2*(c2+1)]) - ( e - Exons[c2] )  +1
							break
					
						# motif on exon 1	
						elif s < Exons[c2] and c2 != (len(Exons)-1):
							begin=int(vector[6+2*c2]) - s  +1
							es=int(vector[5+2*c2])
							ss=int(vector[6+2*(c2+1)])
							end=int(vector[6+2*(c2+1)]) - ( e - Exons[c2] )  +1
							break
						elif s < Exons[c2] and c2 == (len(Exons)-1):
							break
					else:
						
						# in case motif is not on exon 1
						if s < (Exons[c2] - motiflength) and c2 != 0:
							begin=int(vector[5+2*c2]) + ( s - Exons[c2 -1] )  -1
							end=int(vector[5+2*c2]) + ( e - Exons[c2 -1] )   -1
							break
						
						# motif on exon 1
						elif s < (Exons[c2] - motiflength):
							begin=int(vector[5+2*c2]) + s  -1
							end=int(vector[5+2*c2]) + e   -1
							break
						
						# motif overlaps a splice site
						# motif not on the first exon or last exon ( makes no sense but apparently happens)
						elif s < Exons[c2] and c2 != 0 and c2 != (len(Exons) -1):
							begin=int(vector[5+2*c2]) + ( s - Exons[c2 -1] )  -1
							es=int(vector[6+2*c2])
							ss=int(vector[5+2*(c2+1)])
							end=int(vector[5+2*(c2+1)]) + ( e - Exons[c2] )  -1
							break
					
						# motif on exon 1	
						elif s < Exons[c2] and c2 != (len(Exons)-1):
							begin=int(vector[5+2*c2]) + s  -1
							es=int(vector[6+2*c2])
							ss=int(vector[5+2*(c2+1)])
							end=int(vector[5+2*(c2+1)]) + ( e - Exons[c2] )  -1
							break
						elif s < Exons[c2] and c2 == (len(Exons)-1):
							break
								
				# some motifs have coordinates that aren't even on the gene	
				if begin != 0 or end != 0:	
					if es != 0 or ss != 0:
						ID=vector[0].split("|")
						
						# sort the coordinates
						if begin >= end:
							helper = begin
							begin=end
							end=helper
						if es >= ss:
							helper = es
							es = ss
							ss = helper
						
						# print output	
						print ( ID[1] + "\t" + vector[3] + "\t"  + "%d" % begin + "\t"  + "%d" % es + "\t" + "%d" % ss + "\t" + "%d" % end )	
					else:
						ID=vector[0].split("|")
						
						# sort the coordinates
						if begin >= end:
							helper = begin
							begin=end
							end=helper
							
						# print output
						print ( ID[1] + "\t" + vector[3] + "\t"  + "%d" % begin + "\t" + "%d" % end )

# close files
try:
	transcripts.close()
	archive.close()
except:
	pass
motif.close()
