# Description: This script is used to introduce T>C conversions
#              to the filtered reads. Therefore, the motif 
#              patterns are individually searched and their location
#              stored. Subsequently, the reads are mutated according
#              to an array of conversion rates.

# Contact: wanja.kassuhn@web.de

# import
import sys
import os
import random
import re
import math
from collections import defaultdict


############IMPORTANT##################
motifStart = 31           # line in the mutation rate file in which the motif starts
motifStartTrunc=61
maxreadlength=50          # is involved in deciding if a read is spliced
ConversionProb = [0.75,0.33] # length of this array decides the maximum amount of conversions (floats set the amount of reads of the previous round to be edited)
DeletionProb = [0.001]
#######################################

# init
numberReads=0
numberOfSpliced=0
numberOfunpliced=0
dict = defaultdict(list)
splicejunctions = defaultdict(list)
line2=""
motiflength=int(sys.argv[8])
mode = sys.argv[10]
Boundaries = sys.argv[11]
BoundariesGiven = True
lower = 0
upper = 0
activeConv = False
activeDel = False
activeTrunc = False

#init Trunc
numberOfTrunc=0
isTrunc = False
truncLeft=-1
ratesTrunc=[]
truncDistGiven = False
truncLen=[]

#init Conv
numberMut = 0
numberOfMut0=0
numberOfMut1=0
numberOfMut2=0
numberOfMut3over=0
ratesConv=[]

#init Del
numberOfDel=0
posOfDel=-1
ratesDel=[]

#open files
fasta=open(sys.argv[6],"w")
fasta2=open(sys.argv[7], "w")
reads=open(sys.argv[1] ,"r")
motifs = open(sys.argv[5], "r")

####MODES####
#decide which operations/diagnostic events are carried out
try:
	indMode = mode.split(',')
except:
	sys.stderr.write('ERROR: Something is wrong with the entered mutation mode "%s".\n' % mode)

try:
	lower = int(Boundaries.split(",")[0])
	upper = int(Boundaries.split(",")[1])
except:
	BoundariesGiven=False
	
#TRUNCATION DISTRIBUTION
if not sys.argv[12][-1:] == "-":
	truncs = open(sys.argv[12],'r')
	truncDistGiven = True
	for line in truncs:
		if not line.startswith("#"):
			truncLen.append(line[:-1])

for i in range(len(indMode)):
	if indMode[i] == 'C':
		activeConv = True
	if indMode[i] == 'D':
		activeDel = True
	if indMode[i] == 'T':
		activeTrunc = True

####MOTIFS####
# create dictionary of RBP binding sites
for line in motifs:		
	# skip first line
	if not line.startswith('#'):
		sep = line.split("\t")
		IDs = sep[0]
		# init
		start=0
		end=0
		# set start and end coordinate of each motif
		if len(sep)==4:
			start=sep[2]
			end=sep[3][:-1]
		elif len(sep)==6:
			start=sep[2]
			end=sep[5][:-1]
		else:
			continue
		# fill dictionary
		if IDs in dict:
			dict[IDs].append(start)
			dict[IDs].append(end)
		else:	
			dict[IDs] = [start,end]
# close file
motifs.close()

####READ ANNOTATION####
if activeTrunc:
	# read in gtf file
	annotation = open(sys.argv[9], "r")	
	# iterate gtf file
	for line in annotation:
		if not line.startswith("#"):
			if line.split("\t")[2] == "exon":
				# generate dictionary of transcript IDs and corresponding exon coordinates
				if line[:-1].split("\t")[8].split(";")[1].split(" ")[1] == "transcript_id":
					annotatedTranscripts = line[:-1].split("\t")[8].split(";")[1].split(" ")[2].split(".")[0].strip('"')
					exonstart = int(line.split("\t")[3])
					exonend = int(line.split("\t")[4])
					
					splicejunctions[annotatedTranscripts].append(exonstart)
					splicejunctions[annotatedTranscripts].append(exonend)
	annotation.close()

####READ CONVRATES####
if activeConv:
	try:
		ratesF=open(sys.argv[2],"r")
		# create an array of mutation rates
		for line in ratesF:
			if not line.startswith('#'):
				ratesConv.append(float(line[:-1]))
		ratesF.close()
	except:
		print "ERROR: Could not find conversion rate file."
####READ DELRATES####
if activeDel:
	try:
		ratesF=open(sys.argv[3],"r")
		# create an array of mutation rates
		for line in ratesF:
			if not line.startswith('#'):
				ratesDel.append(float(line[:-1]))
		ratesF.close()
	except:
		print "ERROR: Could not find deletion rate file."
####READ TRUNCRATES####
if activeTrunc:
	try:
		ratesF=open(sys.argv[4],"r")
		# create an array of mutation rates
		for line in ratesF:
			if not line.startswith('#'):
				ratesTrunc.append(float(line[:-1]))
		ratesF.close()
	except:
		print "ERROR: Could not find truncation rate file."
####MUTATION####
# iterate over reads
for line in reads:
	# headers 
	if line.startswith(">"):
		# store headers
		line2=line
	else:
		length=len(line)-1
		
		# discard too short reads
		if length < motiflength:
			continue
			
		# variables
		start=-100
		end=-100
		
		# split the header to get coordinates and transcript ID
		header = line2.split(':')
		
		# extract information from read header
		transcriptID = header[2].split('.')[0]
		readstart = int(header[8].split('-')[0])
		readend = int(header[8].split('-')[1][:-2])
		direction = header[8].split('-')[1][-2:-1]
		
		mutated=list(line[:-1])
		readlength = len(mutated)
		
		if transcriptID in dict:
			
			# check all motifs on current transcript
			for entry in range(0,len(dict[transcriptID]),2):
				
				# check which motif is overlapped by read
				if readstart < int(dict[transcriptID][entry]) and readend > int(dict[transcriptID][entry+1]):
					
					# set motif start and end
					start = int(dict[transcriptID][entry]) - int(readstart) -1 
					end = int(dict[transcriptID][entry+1]) - int(readstart)
					
					# if there is a splice junction in the motif it will be found
					# however, a splice junction in front of the motif is problematic
					if readend - readstart > maxreadlength:
						#(splice site in front of motif)
						start = int(dict[transcriptID][entry]) - int(readstart) -1
						end = start + motiflength
						# debugging
						#print str(start) + '\t' + dict[transcriptID][entry] + '\t' + dict[transcriptID][entry+1] + '\t' + str(end)
						if not start <= length - motiflength or not start == end - motiflength or not end >= start + motiflength:
							#(splice site after motif)
							end = int(readend) - int(dict[transcriptID][entry+1])
							start = length - end - (int(dict[transcriptID][entry+1])-(int(dict[transcriptID][entry])-1))
					# debugging
					#print str(start) + '\t' + dict[transcriptID][entry] + '\t' + dict[transcriptID][entry+1] + '\t' + str(end)
		
				# depending on read direction motif coordinates need to be counted from the right instead of from the left		
				if direction == "C":
					end = length - start
					start = end - motiflength
		
		# debugging
		#print str(start) + '\t' + str(end)	
		
		# reads without motif		
		if (start == -100 and end == -100) or (start > length - motiflength or start != end - motiflength):
			continue
		
		if start != -100 and end != -100:
			
			# statistics
			if not activeTrunc:
				numberReads+=1 
			
			####TRUNCATIONS####
			if activeTrunc:
				n = len(mutated)
				SUM=0.0
				border=0.0
				for i in range(n):
					try:
						SUM+=ratesTrunc[motifStartTrunc - start + i]
					except:
						sys.stderr.write("ERROR: The read length exceeds the boundaries of the provided truncation rate file: %s. You should add more lines or reduce the read length!" % sys.argv[4])
						sys.exit()
					
				prob = random.random() * SUM
				for i in range(n):
					border+=ratesTrunc[motifStartTrunc - start + i]
					if prob <= border:
						truncLeft=i
						break
					else:
						continue
				
				# GET TRUNCATED SPLICE JUNCTIONS	
				coordinates = []
				if transcriptID in splicejunctions:
					coordinates = sorted(splicejunctions[transcriptID])
				exons = len(coordinates)
				for i in range(1,exons-1,2):
					# scan for reads overlapping splice junction					
					if int(coordinates[i]) >= readstart and int(coordinates[i+1]) <= readend:
						######CHECK#####
						if direction == "C":
							if truncLeft <= (readend - coordinates[i+1]):
								readend-=truncLeft
								break
							# (start - coordinate[i+1])
							elif truncLeft >= (readend - coordinates[i+1]):
								readend = coordinates[i] - (truncLeft - (readend - coordinates[i+1]))
								break
						else:
							if truncLeft <= (coordinates[i] - readstart):
								readstart+=truncLeft
								break
							# (start - coordinate[i+1])
							elif truncLeft >= (coordinates[i] - readstart):
								readstart = truncLeft - (coordinates[i] - readstart) + coordinates[i+1]
								break

				else:
					readstart = readstart + truncLeft #+ 1
							
			####CONVERSIONS####
			if activeConv:
				numberMut=0
				
				for c in range(len(ConversionProb)):
					
					# exclude reads in each loop that were not mutated in the previous
					if numberMut < c:
						break
					
					# find all "T" and save their location
					a = [m.start() for m in re.finditer('T', "".join(mutated) )]
					n = len(a)
					SUM = 0.0
					# sum up mutation rates of "T" positions from the mutation rate array for normalization
					for i in range(n):
						SUM+=ratesConv[motifStart - start + a[i]]
					
					# normalize mutation rates
					SUM2 = SUM / ConversionProb[c]
					# generate random float
					prob = random.random() * SUM2
					border=0.0
					
					#for all "T" look if the float is higher than the border
					for i in range(n):
						border+=ratesConv[motifStart - start + a[i]]
						if prob <= border:
							#t->c
							if i > truncLeft:
								mutated[int(a[i])]="C"
								numberMut+=1
							break
						else:
							continue
				
				# count mutations for statistics
				if numberMut == 0:
					numberOfMut0+=1
				elif numberMut == 1:
					numberOfMut1+=1
				elif numberMut == 2:
					numberOfMut2+=1
				else:
					numberOfMut3over+=1
					
			####DELETIONS#####
			if activeDel:
				# find all "T" and save their location
				a = [m.start() for m in re.finditer('T', line )]
				posOfDel=-1
				n = len(a)
				SUM=0.0
				border=0.0
				
				for i in range(n):
					SUM+=ratesDel[motifStart - start + a[i]]
					
				# normalize mutation rates
				SUM2 = SUM / DeletionProb[0]
				
				prob = random.random() * SUM2
				
				for i in range(n):
					border+=ratesDel[motifStart - start + a[i]]
					if prob <= border:
						if i > truncLeft:
							mutated[int(a[i])]=""
							numberOfDel+=1
							posOfDel=a[i]
						break
					else:
						continue
					
				if not activeTrunc:
					if posOfDel == 0:
						readstart += 1
					elif posOfDel > 0:
						readend -= 1
				
				#else: ????? 
				
			
			# reset
			isTrunc = False
			# truncate sequence
			if activeTrunc:
				for i in range(truncLeft):
					mutated[i]=""
				if BoundariesGiven:
					if len("".join(mutated)) < lower or len("".join(mutated)) > upper:
						continue
					else:
						if truncDistGiven:
							isTrunc = True
						else:
							numberOfTrunc+=1
							numberReads+=1
				#discard too short reads
				else:
					if len("".join(mutated)) <= readlength / 2 - readlength / 8 or  len("".join(mutated)) >= readlength * 3 / 4:
						continue
					else:
						if truncDistGiven:
							isTrunc = True
						else:
							numberOfTrunc+=1
							numberReads+=1
					
			if readend - readstart > maxreadlength:
				numberOfSpliced+=1
			else:
				numberOfunpliced+=1
						
			# change end position when read is truncated and add it to the vector that represents the sequence header
			header[8] = str(readstart) +"-"+ str(readend) + direction
			if activeConv:
				header.append(str(numberMut))
			if activeDel:
				header.append(str(posOfDel)) #for deletions
				
			if truncDistGiven:
				if isTrunc:
					#sys.stderr.write( " %s \n" % truncLen)
					rand = random.random()
					#sys.stderr.write( "%f \n" % rand)
					lengthOfTruncRead = len("".join(mutated))
					if rand <= 1.0 - float(truncLen[lengthOfTruncRead]):
						
						numberOfTrunc+=1
						numberReads+=1
						
						# print joined sequence header to stdout		
						print ":".join(header)
						# joining array of sequence header
						line2= ":".join(header)
						# sequence header to fasta
						fasta.write(line2+"\n")
						# mutated sequence to stdout
						print "".join(mutated)
						# unmutated Sequence to fasta file
						fasta.write(line)
					
					else:
						#sys.stderr.write("1")
						continue	
			
			else:
				# print joined sequence header to stdout		
				print ":".join(header)
				# joining array of sequence header
				line2= ":".join(header)
				# sequence header to fasta
				fasta.write(line2+"\n")
				# mutated sequence to stdout
				print "".join(mutated)
				# unmutated Sequence to fasta file
				fasta.write(line)
			
# dataset output
fasta2.write("Datastats"+"\n")
fasta2.write("Data contains %i reads with 0 conversions" % numberOfMut0 + "\n")
fasta2.write("Data contains %i reads with 1 conversions" % numberOfMut1 + "\n")
fasta2.write("Data contains %i reads with 2 conversions" % numberOfMut2 + "\n")
fasta2.write("Data contains %i reads with more that 2 conversions" % numberOfMut3over + "\n")
fasta2.write("Data contains %i reads with a T deletion" % numberOfDel + "\n")
fasta2.write("Data contains %i truncations" % numberOfTrunc + "\n")
fasta2.write("Data contains %i reads" % numberReads + "\n")
fasta2.write("Of those are {0} spliced and {1} unspliced".format(numberOfSpliced , numberOfunpliced ))

# close files
try:
	splicejunctions.clear()
except:
	pass

fasta.close()
fasta2.close()
reads.close()

exit