# Description: This script is used to introduce T>C conversions
#              to the filtered reads. Therefore, the motif 
#              patterns are individually searched and their location
#              stored. Subsequently, the reads are mutated according
#              to an array of conversion rates.

# Contact: wanja.kassuhn@web.de

# import
import sys
import os
import random
import re
import math
from collections import defaultdict


############IMPORTANT##################
motifStart = 31           # line in the mutation rate file in which the motif starts
maxreadlength=50          # is involved in deciding if a read is spliced
MutationProb = [0.5,0.5] # length of this array decides the maximum amount of conversions ( floats influence the mutation rates)
#######################################

# init
dict = defaultdict(list)
line2=""

# initialize variables for mutations
numberMut = 0
numberReads=0
numberOfMut0=0
numberOfMut1=0
numberOfMut2=0
numberOfSpliced=0
numberOfunpliced=0

motiflength=int(sys.argv[8])

# init for truncations
numberOfTrunc=0
truncRight=0
truncLeft=0
truncrate=[0.05,0.2,0.75]

# init for deletions
numberOfDel=0
#deletion rates for "T", "TT", and "TTT" deletions
#deletionRates=[0.2,0.1,0.05]
posOfDel=0
delOptions=[]
dele=0

# store paths for unmutated reads and a statistic output
output1 = sys.argv[6]
output2 = sys.argv[7]

# open index file
motifs = open(sys.argv[5], "r")

# create dictionary of RBP binding sites
for line in motifs:
		
	# skip first line
	if not line.startswith('#'):
		sep = line.split("\t")
		IDs = sep[0]
		
		# init
		start=0
		end=0
		
		# set start and end coordinate of each motif
		if len(sep)==4:
			start=sep[2]
			end=sep[3][:-1]
		elif len(sep)==6:
			start=sep[2]
			end=sep[5][:-1]
		else:
			continue
		
		# fill dictionary
		if IDs in dict:
			dict[IDs].append(start)
			dict[IDs].append(end)
		else:	
			dict[IDs] = [start,end]

# close file
motifs.close()

#decide which operations/diagnostic events are carried out
mode = sys.argv[10]
activeConv = False
activeDel = False
activeTrunc = False
try:
	indMode = mode.split(',')
except:
	sys.stderr.write('ERROR: Something is wrong with the entered mutation mode "%s".\n' % mode)

for i in range(len(indMode)):
	if indMode[i] == 'C':
		activeConv = True
	if indMode[i] == 'D':
		activeDel = True
	if indMode[i] == 'T':
		activeTrunc = True

if activeTrunc:
	# read in gtf file
	splicejunctions = defaultdict(list)
	annotation = open(sys.argv[9], "r")
	
	# iterate gtf file
	for line in annotation:
		
		if not line.startswith("#"):
			
			if line.split("\t")[2] == "exon":
				
				# generate dictionary of transcript IDs and corresponding exon coordinates
				if line[:-1].split("\t")[8].split(";")[1].split(" ")[1] == "transcript_id":
					annotatedTranscripts = line[:-1].split("\t")[8].split(";")[1].split(" ")[2].split(".")[0].strip('"')
					exonstart = int(line.split("\t")[3])
					exonend = int(line.split("\t")[4])
					
					splicejunctions[annotatedTranscripts].append(exonstart)
					splicejunctions[annotatedTranscripts].append(exonend)
					
		
	annotation.close()

# open filtered reads
reads=open(sys.argv[1] ,"r")

# open file with the position specific mutation rates
ratesF=open(sys.argv[2],"r")

# create output files
fasta=open(output1,"w")
fasta2=open(output2, "w")

# create an array of mutation rates
rates=[]
for line in ratesF:
	rates.append(float(line[:-1]))
ratesF.close()

# iterate over reads
for line in reads:
	
	# headers 
	if line.startswith(">"):
		# store headers
		line2=line
	else:
		length=len(line)-1
		
		# discard too short reads
		if length < motiflength:
			continue
			
		# variables
		start=-100
		end=-100
			
		# copy of sequence for mutation
		mutated=list(line[:-1])
		
				
		# split the header to get coordinates and transcript ID
		header = line2.split(':')
		
		# extract information from read header
		transcriptID = header[2].split('.')[0]
		readstart = int(header[8].split('-')[0])
		readend = int(header[8].split('-')[1][:-2])
		direction = header[8].split('-')[1][-2:-1]
		
		if transcriptID in dict:
			
			# check all motifs on current transcript
			for entry in range(0,len(dict[transcriptID]),2):
				
				# check which motif is overlapped by read
				if readstart < int(dict[transcriptID][entry]) and readend > int(dict[transcriptID][entry+1]):
					
					# set motif start and end
					start = int(dict[transcriptID][entry]) - int(readstart) -1 
					end = int(dict[transcriptID][entry+1]) - int(readstart)
					
					# if there is a splice junction in the motif it will be found
					# however, a splice junction in front of the motif is problematic
					if readend - readstart > maxreadlength:
						
						#print "mark"
						#print length
						
						#(splice site in front of motif)
						start = int(dict[transcriptID][entry]) - int(readstart) -1
						end = start + motiflength
						
						# debugging
						#print str(start) + '\t' + dict[transcriptID][entry] + '\t' + dict[transcriptID][entry+1] + '\t' + str(end)
						
						if not start <= length - motiflength or not start == end - motiflength or not end >= start + motiflength:
						
							#(splice site after motif)
							end = int(readend) - int(dict[transcriptID][entry+1])
							start = length - end - (int(dict[transcriptID][entry+1])-(int(dict[transcriptID][entry])-1))
					
					# debugging
					#print str(start) + '\t' + dict[transcriptID][entry] + '\t' + dict[transcriptID][entry+1] + '\t' + str(end)
		
				# dipending on read direction motif coordinates need to be counted from the right instead of from the left		
				if direction == "C":
					end = length - start
					start = end - motiflength
		
		# debugging
		#print str(start) + '\t' + str(end)	
		
		# reads without motif		
		if (start == -100 and end == -100) or (start > length - motiflength or start != end - motiflength):
			continue
		
		# if read contains a motif		
		if start != -100 and end != -100:
			
			# statistics
			numberReads+=1 
			
			if activeConv:

				# mutation
				# find all "T" and save their location
				a = [m.start() for m in re.finditer('T', line )]
				n = len(a)
				SUM = 0.0
				numberMut=0
				# sum up mutation rates of "T" positions from the mutation rate array for normalization
				for i in range(n):
					SUM+=rates[motifStart - start + a[i]]
				
				
				for c in range(len(MutationProb)):
					# normalize mutation rates
					SUM2 = SUM / MutationProb[c]
					# generate random float
					i = random.random() * SUM2
					border=0.0
					#for all "T" look if the float is higher than the border
					for c2 in range(n):
						border+=rates[motifStart - start + a[c2]]
						if i <= border:
							#t->c
							mutated[int(a[c2])]="C"
							numberMut+=1
							break
						else:
							continue
			
			# debugging
			#print str(start) + '\t' + str(int(dict[transcriptID][entry+1])-(int(dict[transcriptID][entry])-1)) + '\t' + str(end)
			
#			#Deletions
#			posOfDel=0
#			delOptions=[]
#			dele=0
#			i = random.random()
#			#get all T in front of the motif 
#			for c in range(n):
#				if a[c] < start:
#					delOptions.append(a[c])
#					
#			#each t has same chance to be deleted			
#			p = 1/len(delOptions)
#			SUM=0
#			
#			for c in range(len(delOptions)):
#				SUM+=p
#				#pick a position randomly
#				if i <= SUM:
#					# delete with a 1 in 999 chance
#					i2 = random.randint(1,1000)
#					if i2 == 500:
#						mutated[a[c]] = ""
#						dele=1
#						posOfDel=delOptions[c]
#						numberOfDel+=1
#						break
#			
#
#			#choose t to be deleted
#			for c in range(n):
#				SUM+=p
#				if i <= SUM:
#					if a[c-1] == a[c]-1 and a[c+1] == a[c]+1:
#						posOfDel=a[c]-1
#					if a[c-1] == a[c]-1:
#						posOfDel=a[c]-1
#					if a[c+1] == a[c]+1:
#						posOfDel=a[c]
#					else:
#						posOfDel=a[c]
#						
#					i2 = random.random()
#					
#					mutated[a[c]] = ""
#					numberOfDel+=1
#					continue
#					
#
#		 	
			if activeTrunc:
				# initiate truncations
				truncRight=0
				truncLeft=0
				
				# truncations
				# find all "G" and save their location in an array
				a = [m.start() for m in re.finditer('G', line )]
				
				
				# generate normalized random float
				Sum=0.0
				for i in range(len(truncrate)):
					Sum += truncrate[i]
				i = random.random() * Sum
				
				# count trancation options in front and after the the motif
				posLeft = sum(c < (start - 4) for c in a)
				posRight = sum(c > (end + 4) for c in a)
				
				# set counter for truncation array !!! must be (len(truncrate)-1)
				counter = 2
				
				# for each of the "G" behind the motif generate a random float, look in the truncation rate array if the float is higher than the entry
				# count up the truncation options and continue
				for c in range(len(a)):
					
					# create a border around the motif of 4 nt and add trancation upstream
					if a[c] < (start-4):
						try:
							
							if i <= truncrate[c - posLeft + len(truncrate)]:
								#count the truncation for statistics
								numberOfTrunc+=1  # add boolean to avoid counting twice
								# truncation is at position: sequence length - position of truncated "G"
								truncLeft = a[c] +1
								continue
						except:
							continue
						
					# create a border around the motif of 4 nt and add trancation downstream		
					elif a[c] > (end+4):
						try:			
							# trunc rate array not long enough, when surpassed mutate with 100%
							if i <= truncrate[counter]:
								#count the truncation for statistics
								numberOfTrunc+=1   # add boolean to avoid counting twice
								# truncation is at position: sequence length - position of truncated "G"
								truncRight = len(mutated) - (a[c]+1)
								continue
							else:
								counter-=1
						except:
							break
				
				# truncate the array that includes the mutated sequence by trunc
				if truncRight > 0:
					for c in range(truncRight):
						mutated[len(mutated)-(c+1)]=""
						
				# truncate the array that includes the mutated sequence by trunc
				if truncLeft > 0:
					for c in range(truncLeft):
						mutated[c]=""
			
			# count mutations for statistics
			if numberMut == 0:
				numberOfMut0+=1
			elif numberMut == 1:
				numberOfMut1+=1
			else:
				numberOfMut2+=1
			
			if activeTrunc:
				# generate splice junctions
				coordinates = []
				if transcriptID in splicejunctions:
					coordinates = sorted(splicejunctions[transcriptID])
				exons = len(coordinates)
				
				# debugging
				#print coordinates
				
				for i in range(1,exons-1,2):
					# scan for reads overlapping splice junction
					if coordinates[i] > readstart and coordinates[i+1] < readend:
						
						# debugging
						#print str(readstart) + "\t" + str(readend)
						#print str(truncLeft) + "\t" + str(truncRight)
						#print str(i) + "\t" + str(i+1)
						
						# if truncation cuts away a splice junction, add intron length to truncation length
						if readstart + truncLeft > coordinates[i]:
							truncLeft = truncLeft + coordinates[i+1] - coordinates[i]
						elif readend - truncRight < coordinates[i+1]:
							truncRight = truncRight + coordinates[i+1] - coordinates[i]
			
			
				
			# look if a read is spliced and count the number of unspliced and spliced reads for statistics
			if readend - readstart > maxreadlength:
				numberOfSpliced+=1
			else:
				numberOfunpliced+=1
				
			if activeTrunc:
				if direction == 'C':
					readstart = readstart + truncRight
					readend = readend - truncLeft
				else:
					readstart = readstart + truncLeft
					readend = readend - truncRight
			
			# change end position when read is truncated and add it to the vector that represents the sequence header
			header[8] = str(readstart) +"-"+ str(readend) + direction
			header.append(str(numberMut))
			#vec.append(str(posOfDel)) #for deletions
			
			# print joined sequence header to stdout		
			print ":".join(header)
			# joining array of sequence header
			line2= ":".join(header)
			# sequence header to fasta
			fasta.write(line2+"\n")
			# mutated sequence to stdout
			print "".join(mutated)
			# unmutated Sequence to fasta file
			fasta.write(line)
			
# dataset output
fasta2.write("Datastats"+"\n")
fasta2.write("Data contains %i reads with 0 conversions" % numberOfMut0 + "\n")
fasta2.write("Data contains %i reads with 1 conversions" % numberOfMut1 + "\n")
fasta2.write("Data contains %i reads with 2 conversions" % numberOfMut2 + "\n")
fasta2.write("Data contains %i reads with a T deletion" % numberOfDel + "\n")
fasta2.write("Data contains %i truncations" % numberOfTrunc + "\n")
fasta2.write("Data contains %i reads" % numberReads + "\n")
fasta2.write("Of those are {0} spliced and {1} unspliced".format(numberOfSpliced , numberOfunpliced ))

# close files
#splicejunctions.clear()
fasta.close()
fasta2.close()
reads.close()

exit